#include "xbp.ch"
#include "common.ch"
#include "xbzlib.ch"

CLASS XbpBitmap2 FROM XbpBitmap
   PROTECTED:
      VAR aChunk
      VAR aPalette
      METHOD SaveImgAsPNG
      METHOD _IHDR
      METHOD _PLTE
      METHOD _tIME
      METHOD _pHYs
      METHOD _tEXt
      METHOD _zTXt
      METHOD _IDAT
      METHOD _IEND
      METHOD _software
      METHOD makeChunk

   EXPORTED:
      METHOD saveFile
ENDCLASS

METHOD XbpBitmap2:saveFile(cImgFile, nFormat, nCompress, lOverWrite, sComment, lComment)
   local lOk

   default lOverWrite to FALSE

   // copy XbpBitmap behaviour for existing file
   if lOverWrite .and. FExists(cImgFile)
      FErase(cImgFile)
   endif
   
   if nFormat == XBPBMP_FORMAT_PNG
      lOk := ::SaveImgAsPNG( cImgFile, nFormat, nCompress, sComment, lComment )
   else
      lOk := ::XbpBitmap:saveFile(cImgFile, nFormat, nCompress)
   endif
   return lOk

METHOD XbpBitmap2:SaveImgAsPNG( cImgFile, nFormat, nCompress, sComment, lComment )
   local lOk := FALSE
   local aPalette
   local i
   local cBuff := PNG_SIGNATURE

   ::aPalette := ::getColorTable()
   if !FExists(cImgFile) .and. !(::aPalette == NIL)
      ::aChunk := {}
      ::_IHDR(iif(Valtype(::aPalette) == 'A', Len(::aPalette), 255))
      ::_tIME()
      ::_pHYs()
      if !(::aPalette == NIL)
         ::_PLTE(::aPalette)
      endif

      if !(sComment == NIL)
         ::_tEXt(sComment)
      endif

      if !(lComment == NIL)
         ::_zTXt(lComment)
      endif
      ::_Software()

      ::_IDAT(nCompress)
      ::_IEND()

      for i := 1 to Len(::aChunk)
         cBuff += ::aChunk[i]
      next
      i := FCreate(cImgFile)
      if i > 0
         lOk := FWrite( i, cBuff, Len(cBuff) ) == Len(cBuff)
         FClose(i)
      endif
   elseif !FExists(cImgFile)
      lOk := ::XbpBitmap:saveFile(cImgFile, nFormat, nCompress)
   endif
   return lOk


METHOD XbpBitmap2:_IHDR(nColors)
   local x        := N2NBO( U2Bin(::xSize) )
   local y        := N2NBO( U2Bin(::ySize) )
   local bDepth   := Chr( ::bits )
   local cType    := Chr(iif( nColors <= 256, 3, 2 ))
   local Comp     := Chr(0)
   local cFilt    := Chr(0)
   local cIntrl   := Chr(0)

   local cData := x+y+bDepth+;
                  cType+Comp+cFilt+cIntrl

   aadd( ::aChunk, ::makeChunk('IHDR', cData) )

   return self

METHOD XbpBitmap2:_PLTE(aPalette)
   local cBuff := Space( Len(aPalette)*3 )
   local i
   local n

   for i := 1 to len(aPalette)
      for n := 1 to 3
         cBuff[((i*3)-3)+n] := Chr(aPalette[i][n])
      next
   next
   aadd( ::aChunk, ::makeChunk('PLTE',cBuff))

   return self

METHOD XbpBitmap2:_tIME()
   local cYear    := N2NBO( I2Bin( Year(Date()) ) )
   local cMonth   := Chr( Month(Date()) )
   local cDay     := Chr( Day(Date()) )
   local cTime    := StrTran( Time(), ':', '' )
   local cHr      := Chr( Val(SubStr(cTime,1,2)) )
   local cMin     := Chr( Val(SubStr(cTime,3,2)) )
   local cSec     := Chr( Val(SubStr(cTime,5,2)) )
   local cBuff := cYear+cMonth+cDay+cHr+cMin+cSec

   aadd( ::aChunk, ::makeChunk('tIME',cBuff) )
   return self

METHOD XbpBitmap2:_pHYs()
   local cBuff := Chr(0)+Chr(0)+Chr(0x0b)+Chr(0x12)
   cBuff += cBuff+Chr(1)
   aadd( ::aChunk, ::makeChunk('pHYs',cBuff) )
   return self

METHOD XbpBitmap2:_tEXt(sText)
   local cBuff := 'Title'+Chr(0)+sText

   aadd( ::aChunk, ::makeChunk('tEXt',cBuff) )
   return self

METHOD XbpBitmap2:_zTXt(lText)
   local cBuff := 'Description'+Chr(0)+Chr(0)+xbzCompress(lText)+Chr(0)

   aadd( ::aChunk, ::makeChunk('zTXt',cBuff) )
   return self

METHOD XbpBitmap2:_Software()
   local cTxt := 'Xbase++ v'+Version(1)+'.'+Version(2)+'.'+Version(3)+', XbZLib v'+LTrim(Str(XbZLibVer()))+Chr(0)
   local cBuff := 'Software'+Chr(0)+Chr(0)+xbzCompress(cTxt)

   aadd( ::aChunk, ::makeChunk('zTXt',cBuff) )
   return self

METHOD XbpBitmap2:_IDAT(nCompress)
   local cBuff := ::setBuffer()
   local x := ::xSize
   local y := ::ySize
   local i, n, v
   local cLine
   local r, g, b
   local cData := ''

   b := ::bits
   altd()

   // <= 256 colour
   if !(::aPalette == NIL)
      cBuff:= Right( cBuff, x*y )
      for i := 1 to y
         cData += Chr(0)+SubStr(cBuff, ((i*y)-y)+1, x )
      next
   else // true-colour

      x := x*(b/8)
      cBuff:= Right( cBuff, x*y )
      for i := 1 to y
         cData += Chr(0)+SubStr(cBuff, ((i*x)-x)+1, x )
      next

   endif
   cData := xbzCompress2(cData,NIL,nCompress)

   aadd( ::aChunk, ::makeChunk('IDAT',cData) )
   return self

METHOD XbpBitmap2:_IEND()
   aadd( ::aChunk, ::makeChunk('IEND','') )
   return self

METHOD XbpBitmap2:makeChunk(cChunk,cData)
   local cLen := N2NBO(U2Bin(Len(cData)))
   local cBuff := cChunk+cData
   local cCRC  := N2NBO( crc32( cBuff ) )
   return cLen+cBuff+cCrc

STATIC Function N2NBO(c)
   local i
   local cRet := Space(Len(c))

   for i := 1 to Len(c)
      cRet[i] := c[-i]
   next
   return cRet


